/*Author: Jin-Hee Cho 
* Project Name: Wireless Ad Hoc Messenger Using Pocket PCs
* Purpose: Thesis for Master Science in Computer Science 
* at Virginia Polytechnic State University
* Funded by Microsoft
* Date Updated: 7/29/2004
* SendAndReceiveFiles is the application to wirelessly move result files 
* in each Pocket PC to the Statistical Analyzer running on the laptop.
*/

using System;
using System.Drawing;
using System.Collections;
using System.ComponentModel;
using System.Windows.Forms;
using System.Threading;
using System.Net.Sockets;
using System.Net;
using System.IO;
using System.Text;
using System.Xml;

namespace SendAndReceiveFiles
{
	/// <summary>
	/// Summary description for SendAndReceiveFiles.
	/// </summary>
	public class SendAndReceiveFiles : System.Windows.Forms.Form
	{
		private System.Windows.Forms.Button SendFile_Btn;
		private System.Windows.Forms.Label label1;
		private System.Windows.Forms.MainMenu mainMenu1;
		private System.Windows.Forms.Button Exit_Btn;
		private System.Windows.Forms.TextBox MyID_IP_TB;
		private System.Windows.Forms.ListBox SendFile_LB;
	
		//***********************************************************************//
		//UdpServer
		//***********************************************************************//
		private Socket udpSocket;
		private Thread udpListeningThread;
		public string ReceivedData;
		public int ReceivedDataLength;
	
		//********************************************************************//
		//UdpClient
		private UdpClient udpClient;		
		private bool udpServerIsAlive;

		//Other variables
		private string strMyIP;
		private string strMyID;
		private string ResultsCollected;
		private int fileID;
		public Hashtable SenderBuffer;
		private ArrayList DataToSendBuffer;
		private Hashtable ReceiptTimerBuffer;
		private string file_name;
		private XmlDocument xmlDoc;		
		string received_file_name = "";
		

		public SendAndReceiveFiles()
		{
			//
			// Required for Windows Form Designer support
			//
			InitializeComponent();

			//
			// TODO: Add any constructor code after InitializeComponent call
			//
			try
			{
				fileID  = -1;
				ResultsCollected ="";				
				udpServerIsAlive = false;
				SenderBuffer = new Hashtable();
				start_server();												
				DataToSendBuffer = new ArrayList();
				ReceiptTimerBuffer = new Hashtable();
				xmlDoc = new XmlDocument();
				xmlDoc.Load(new StreamReader("\\My Documents\\Personal\\conf.xml"));
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in Constructor(): "+e.Message);
			}

		}

		private void resetAll()
		{
			ReceiptTimerBuffer.Clear();
			DataToSendBuffer.Clear();
			ResultsCollected ="";			
			SenderBuffer.Clear();
			fileID = -1;			
		}

		public void stop_server()
		{
			try
			{	
				if (udpSocket != null)
				{
					udpSocket.Close();
				}
				udpListeningThread=null;//Free Mem
			}
			catch (Exception e)
			{
				MessageBox.Show("stop_server() exception occurred :"+e.Message);
			}
		}		
		private void start_server()
		{
			try
			{		
				//Starting the UDP Listening thread.
				udpListeningThread = new Thread(new ThreadStart(this.receivePacket));
				udpServerIsAlive = true;				
				udpListeningThread.Start();
			}
			catch (Exception e)
			{
				MessageBox.Show("start_server() exception occurred : "+e.Message);
				stop_server();
			}
		}
						
		/// <summary>
		/// Receive UDP Packets
		/// </summary>
		public void receivePacket()
		{			
			IPHostEntry localHostEntry;
		
			try
			{
				udpSocket = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
				localHostEntry = Dns.GetHostByName(Dns.GetHostName());
				IPEndPoint localIpEndPoint = new IPEndPoint(localHostEntry.AddressList[0], UDPCONSTANTS.UDPPORT);
				EndPoint localEndEntry = (localIpEndPoint);
				udpSocket.Bind(localEndEntry);
				while (udpServerIsAlive)
				{
					Monitor.Enter(this);
					Byte[] received = new Byte[UDPCONSTANTS.MESSAGESIZE];
					//This is just a dummy endpoint that will be replaced after the call to ReceiveFrom.
					//In the ReceiveFrom call later on, dummyIpEndPoint will be passed by ref.
					IPEndPoint dummyIpEndPoint = new IPEndPoint(localHostEntry.AddressList[0], UDPCONSTANTS.UDPPORT);

					EndPoint remoteNotifyEP = (dummyIpEndPoint);

					int udpBytesReceived = udpSocket.ReceiveFrom(received, ref remoteNotifyEP);

					ReceivedDataLength=udpBytesReceived;								
					String dataReceivedUdp = System.Text.Encoding.ASCII.GetString(received,0,received.Length );
					ReceivedData=dataReceivedUdp.ToString();								
					HandleReceivedPacket();
					Monitor.Exit(this);		
				}		
					
			}	
			catch (Exception e)
			{
				//MessageBox.Show("Exception occurred in receivedPacket() :"+e.Message);
			}
		}
		
		
		private int getMessageLength()
		{
			string msgLength= "";
			try
			{								
				for(int i =0 ; i<ReceivedData.Length; i++)
				{
					if(ReceivedData[i].Equals(' '))
					{
						break;
					}
					else
					{
						msgLength = msgLength+ReceivedData[i].ToString();
					}
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in getMessageLength(): "+e.Message);
			}
			return int.Parse(msgLength);
		}
		private FilePacket transformToFilePacket()
		{
			FilePacket fp = new FilePacket();
			try
			{
				XmlDocument doc = new XmlDocument();
				doc.LoadXml(ReceivedData);
				ReceivedDataLength=0;//reset the received data's length
				XmlNode root = doc.DocumentElement; 
				XmlNodeList xNodes = root.ChildNodes;
			
				foreach(XmlNode childNode in xNodes)
				{
					
					XmlElement elem = (XmlElement)childNode;
					if(elem.Name.Equals(COMMONCONSTANTS.SENDERIP))
					{
						fp.strSenderIP = elem.FirstChild.Value;
					}
					else if(elem.Name.Equals(COMMONCONSTANTS.ID))
					{
						fp.strID = elem.FirstChild.Value;
					}
					else if(elem.Name.Equals(COMMONCONSTANTS.NUMOFTOTALPACKETS))
					{
						fp.strNumOfTotalPackets= elem.FirstChild.Value;
					}
					else if(elem.Name.Equals(COMMONCONSTANTS.FILECONTENT))
					{
						fp.strXMLFile =  elem.FirstChild.Value;
					}
					else if(elem.Name.Equals(COMMONCONSTANTS.TYPE))
					{
						fp.strType =  elem.FirstChild.Value;					
					}
				}	
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in transformToFilePacket(): "+e.Message);
			}
			return fp;
		}

		private void SetReceiptTimer(FilePacket p_fp)
		{
			try
			{
				MyTimer myTimer = new MyTimer(1000,true,this,p_fp.strID, p_fp);
				ReceiptTimerBuffer.Add(p_fp.strID, myTimer);
				myTimer.SetReceiptTimer();
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in SetReceiptTimer(): "+e.Message);
			}
		}
		private void SendReceiptPacket(ref FilePacket p_fp)
		{
			try
			{
				string targetIP = p_fp.strSenderIP;
				FilePacket fp = new FilePacket();
				fp.strType = COMMONCONSTANTS.RECEIPT_PACKET;
				fp.strSenderIP = strMyIP;
				fp.strID = p_fp.strID;
				fp.strNumOfTotalPackets = COMMONCONSTANTS.NOT_AVAILABLE;
				fp.strXMLFile = COMMONCONSTANTS.NOT_AVAILABLE;

				XmlDocument doc = new XmlDocument();
				doc.LoadXml("<FilePacket></FilePacket>");

				XmlNode root = doc.DocumentElement;

				//1. Type
				XmlElement elem = doc.CreateElement(COMMONCONSTANTS.TYPE);
				elem.InnerText = fp.strType;
				root.AppendChild(elem);

				//2.ID
				elem = doc.CreateElement(COMMONCONSTANTS.ID);
				elem.InnerText = fp.strID;
				root.AppendChild(elem);

				//3. SenderIP
				elem = doc.CreateElement(COMMONCONSTANTS.SENDERIP);
				elem.InnerText = fp.strSenderIP;
				root.AppendChild(elem);

				//4. Number of Total Packets
				elem = doc.CreateElement(COMMONCONSTANTS.NUMOFTOTALPACKETS);
				elem.InnerText = fp.strNumOfTotalPackets;
				root.AppendChild(elem);

				//5. File content
				elem = doc.CreateElement(COMMONCONSTANTS.FILECONTENT);
				elem.InnerText = fp.strXMLFile;
				root.AppendChild(elem);

				string messageToSend = doc.InnerXml.ToString();
			
				sendUDP(targetIP,messageToSend);
 
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in SendReceiptPacket(): "+e.Message);
			}
		}

		private int getSequenceID(string p_ID)
		{
			int SID = -1;
			try
			{
				for(int i = 0; i<p_ID.Length; i++)
				{
					if(p_ID[i].ToString().Equals(":"))
					{
						string sequenceID = p_ID.Substring(i+1);
						SID = int.Parse(sequenceID);
						break;
					}
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in getSequenceID(): "+e.Message);
			}
			return SID;
		}

		private void writeToFile(FilePacket p_fp)
		{
			int index = getSequenceID(p_fp.strID);
			int numOfTotalPackets = int.Parse(p_fp.strNumOfTotalPackets);			
			if(index == 0)
			{
				 received_file_name = p_fp.strXMLFile;
			}
			else if(index == numOfTotalPackets)
			{
				ResultsCollected += p_fp.strXMLFile;
				StreamWriter myWriter = new StreamWriter
					(@"\My Documents\Personal\"+received_file_name,false,Encoding.ASCII);
				myWriter.Write(ResultsCollected);
				myWriter.Close();				
				//Status_TB.Text += "\r\n"+"Completed to Receive "+received_file_name+".";
				resetAll();
			}
			else
			{
				ResultsCollected += p_fp.strXMLFile;			
			}
		}

		private void HandleReceivedPacket()
		{
			try
			{
				if(ReceivedData.Length>0)
				{
					FilePacket fp = transformToFilePacket();
					if(fp.strType.Equals(COMMONCONSTANTS.RECEIPT_PACKET))
					{
			//			MessageBox.Show(fp.strID+"'s receipt received");
						ReleaseReceiptTimer(fp.strID);
						SenderBuffer.Remove(fp.strID);						
						DataToSendBuffer.RemoveAt(0);
						if((DataToSendBuffer.Count!=0) && (SenderBuffer.Count == 0))
						{
							FilePacket temp_fp = new FilePacket();
							temp_fp = (FilePacket)DataToSendBuffer[0];
							SaveInSenderBuffer(temp_fp);
							SendFilePacket(temp_fp);  
						}
						else if(DataToSendBuffer.Count==0)
						{
							//Status_TB.Text += "\r\n"+"Completed to Send "+file_name+".";
						}
					}
					else//File Packet
					{
						SendReceiptPacket(ref fp);
						writeToFile(fp);
					}
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in HandleReceivedPacket(): "+e.Message);
			}
		}
		//UdpClient relate methods
		public bool sendUDP(string serverIPaddr, string p_message)
		{			
			bool sentSuccess = false;
			try
			{
				IPAddress serverIP = IPAddress.Parse(serverIPaddr);
				udpClient = new UdpClient();
				udpClient.Connect(serverIP,UDPCONSTANTS.UDPPORT);
		
				Byte[] inputToBeSent = new Byte[UDPCONSTANTS.MESSAGESIZE];
				//Convert the "message" string to a byte array, and then send it to
				//the remote UDP server using udpClient.
				inputToBeSent = System.Text.Encoding.ASCII.GetBytes(p_message.ToCharArray());
				int nBytesSent = udpClient.Send(inputToBeSent, inputToBeSent.Length);
				sentSuccess = true;
				
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception is occurred in sendUDP() : "+e.Message);				
			}
			udpClient.Close();
			return sentSuccess;
		}

		/// <summary>
		/// Clean up any resources being used.
		/// </summary>
		protected override void Dispose( bool disposing )
		{
			base.Dispose( disposing );
		}

		static void Main(string[] args)
		{		
			Application.Run(new SendAndReceiveFiles());
			//Application.Exit();
		}

		#region Windows Form Designer generated code
		/// <summary>
		/// Required method for Designer support - do not modify
		/// the contents of this method with the code editor.
		/// </summary>
		private void InitializeComponent()
		{
			this.SendFile_Btn = new System.Windows.Forms.Button();
			this.MyID_IP_TB = new System.Windows.Forms.TextBox();
			this.label1 = new System.Windows.Forms.Label();
			this.mainMenu1 = new System.Windows.Forms.MainMenu();
			this.Exit_Btn = new System.Windows.Forms.Button();
			this.SendFile_LB = new System.Windows.Forms.ListBox();
			// 
			// SendFile_Btn
			// 
			this.SendFile_Btn.Location = new System.Drawing.Point(24, 56);
			this.SendFile_Btn.Size = new System.Drawing.Size(200, 24);
			this.SendFile_Btn.Text = "Send File";
			this.SendFile_Btn.Click += new System.EventHandler(this.SendFile_Btn_Click);
			// 
			// MyID_IP_TB
			// 
			this.MyID_IP_TB.Location = new System.Drawing.Point(112, 8);
			this.MyID_IP_TB.Size = new System.Drawing.Size(120, 22);
			this.MyID_IP_TB.Text = "";
			// 
			// label1
			// 
			this.label1.Location = new System.Drawing.Point(24, 8);
			this.label1.Size = new System.Drawing.Size(88, 24);
			this.label1.Text = "My ID and IP :";
			// 
			// Exit_Btn
			// 
			this.Exit_Btn.Location = new System.Drawing.Point(24, 224);
			this.Exit_Btn.Size = new System.Drawing.Size(200, 32);
			this.Exit_Btn.Text = "Exit";
			this.Exit_Btn.Click += new System.EventHandler(this.Exit_Btn_Click);
			// 
			// SendFile_LB
			// 
			this.SendFile_LB.Location = new System.Drawing.Point(24, 104);
			this.SendFile_LB.Size = new System.Drawing.Size(200, 100);
			this.SendFile_LB.SelectedIndexChanged += new System.EventHandler(this.SendFile_LB_SelectedIndexChanged);
			// 
			// SendAndReceiveFiles
			// 
			this.BackColor = System.Drawing.Color.LightSteelBlue;
			this.Controls.Add(this.SendFile_LB);
			this.Controls.Add(this.Exit_Btn);
			this.Controls.Add(this.label1);
			this.Controls.Add(this.MyID_IP_TB);
			this.Controls.Add(this.SendFile_Btn);
			this.Menu = this.mainMenu1;
			this.Text = "SendAndReceiveFiles";
			this.Load += new System.EventHandler(this.SendAndReceiveFiles_Load);

		}
		#endregion

		private void SendAndReceiveFiles_Load(object sender, System.EventArgs e)
		{			
			IPHostEntry localHostEntry = Dns.GetHostByName(Dns.GetHostName());
			IPAddress myIP = localHostEntry.AddressList[0];
			strMyIP = myIP.ToString();	
			strMyID = FindIDbyIPAddressInConfFile(strMyIP);
			MyID_IP_TB.Text = strMyID+" :"+strMyIP;						
			SendFile_LB.Items.Add("Result_"+strMyID+".xml");			
			SendFile_LB.Items.Add("GEN_basicInfo.xml");
			SendFile_LB.Items.Add("conf.xml");
			SendFile_LB.Items.Add("DataAnalysis_"+strMyID+".xml");
			if(isSourceNode())
			{
				SendFile_LB.Items.Add("ADC_basicInfo.xml");
			}
		}

		private bool isSourceNode()
		{
			try
			{
				string path = "\\My Documents\\Personal\\ADC_basicInfo.xml";
				if(File.Exists(path))
				{
					return true;
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in isSourceNode(): "+e.Message);
			}
			return false;
		}

		public string FindIDbyIPAddressInConfFile(string strNodeIPAddress)
		{			
			string strReturnID="NA";
			try
			{
				XmlNode root = xmlDoc.DocumentElement;//configuration root element
				XmlNodeList xNodes = root.ChildNodes;
											
				foreach(XmlNode childNode in xNodes)
				{
					
					XmlElement currentElement = (XmlElement)childNode;
					if(currentElement.GetAttribute("IP").Equals(strNodeIPAddress))
					{
						strReturnID=currentElement.GetAttribute("ID").ToString();
					}
				}									
			}
			catch(Exception ex)
			{
				MessageBox.Show("FindIDbyIPAddress: An Exception has occured. \r\n" + ex.ToString()+ex.Message.ToString(), 
					"Pocket Chat",MessageBoxButtons.OK , MessageBoxIcon.Hand ,MessageBoxDefaultButton.Button1);
			}
			return strReturnID;
		}

		private void DivideIntoProperDataSize(string p_message)
		{
			try
			{
				int numOfTotalPackets = 0;
				int messageSize =  UDPCONSTANTS.MESSAGESIZE-300;
				int round  = 0;
				int wholeMSGLength = p_message.Length;
				double db_messageSize = messageSize;
				double db_wholeMSGLength = wholeMSGLength;
				int int_NOP = (wholeMSGLength/messageSize);
				double db_NOP = (db_wholeMSGLength/db_messageSize);
				if(db_NOP.CompareTo(double.Parse(int_NOP.ToString()))>0)
				{
					numOfTotalPackets = int_NOP+1;
				}
				else
				{
					numOfTotalPackets = int_NOP;				 
				}
				
				//MessageBox.Show("numOfTotalPackets: "+numOfTotalPackets);
				FilePacket initial_fp = new FilePacket();				
				initial_fp = createFilePacket(file_name,numOfTotalPackets.ToString());
				DataToSendBuffer.Add(initial_fp); 		
				while(messageSize*round<wholeMSGLength)
				{
					round++;
					string messageToSend = "";
					if(messageSize*round>wholeMSGLength)
					{
						messageToSend = p_message.Substring(messageSize*(round-1));
					}
					else
					{
						messageToSend = p_message.Substring(messageSize*(round-1),messageSize);
					}					
					FilePacket fp = new FilePacket();
					fp = createFilePacket(messageToSend,numOfTotalPackets.ToString());
					DataToSendBuffer.Add(fp); 			
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in DivideIntoProperDataSize(): "+e.Message);
			}
		}

		private string showDataToSendBuffer()
		{
			string inside = "DataToSendBuffer: ";
			try
			{
				for(int i =0; i<DataToSendBuffer.Count; i++)
				{
					FilePacket fp = (FilePacket)DataToSendBuffer[i];
					inside += fp.strID+" ";
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in showDataToSendBuffer() : "+e.Message);
			}
			return inside;
		}

		private void SendFile_Btn_Click(object sender, System.EventArgs e)
		{
			try
			{			
				resetAll();		
				file_name = SendFile_LB.Items[SendFile_LB.SelectedIndex].ToString();
				StreamReader myReader = new StreamReader(@"\My Documents\Personal\"+file_name);
				string message = myReader.ReadToEnd().ToString();		
				DivideIntoProperDataSize(message);
				FilePacket fp = new FilePacket();
				fp = (FilePacket)DataToSendBuffer[0];
				SaveInSenderBuffer(fp);
				SendFilePacket(fp);
				
			}
			catch(Exception ex)
			{
				MessageBox.Show("Exception in SendFile_Btn_Click() :"+ex.Message);
			}

		}		
		        public void SendFilePacket(FilePacket fp)
				{
					try
					{
						XmlDocument doc = new XmlDocument();
						doc.LoadXml("<FilePacket></FilePacket>");

						XmlNode root = doc.DocumentElement;

						//1. ID
						XmlElement elem = doc.CreateElement(COMMONCONSTANTS.ID);
						elem.InnerText = fp.strID;
						root.AppendChild(elem);

						//2. SenderIP
						elem = doc.CreateElement(COMMONCONSTANTS.SENDERIP);
						elem.InnerText = fp.strSenderIP;
						root.AppendChild(elem);

						//3. Length
						elem = doc.CreateElement(COMMONCONSTANTS.NUMOFTOTALPACKETS);
						elem.InnerText = fp.strNumOfTotalPackets;
						root.AppendChild(elem);

						//4. File Content
						elem = doc.CreateElement(COMMONCONSTANTS.FILECONTENT);
						elem.InnerText = fp.strXMLFile;
						root.AppendChild(elem);

						//5. Type
						elem = doc.CreateElement(COMMONCONSTANTS.TYPE);
						elem.InnerText = fp.strType;
						root.AppendChild(elem);

						string messageToSend = doc.InnerXml.ToString();
				
						SetReceiptTimer(fp);
						bool sentSucess = sendUDP(COMMONCONSTANTS.SERVERIP, messageToSend);
					}
					catch(Exception e)
					{
						MessageBox.Show("Exception in SendFilePacket(): "+e.Message);
					}
				}

		private void SaveInSenderBuffer(FilePacket p_fp)
		{
			SenderBuffer.Add(p_fp.strID, p_fp);
		}

		private string getID()
		{
			fileID++;
			return fileID.ToString();
		}

		private FilePacket createFilePacket(string p_messageToSend, string p_numOfTotalPackets)
		{
			FilePacket fp = new FilePacket();
			try
			{                
				fp.strType = COMMONCONSTANTS.FILE_PACKET;
				fp.strID = strMyID+":"+getID();
				fp.strNumOfTotalPackets = p_numOfTotalPackets;
				fp.strSenderIP = strMyIP;
				fp.strXMLFile = p_messageToSend;
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in createFilePacket(): "+e.Message);
			}
			return fp;
		}

		private void Reset_Btn_Click(object sender, System.EventArgs e)
		{
			resetAll();
		}

		public void ReleaseReceiptTimer(string p_ID)
		{
			try
			{
				if(ReceiptTimerBuffer.Contains(p_ID))
				{
					IDictionaryEnumerator ide = ReceiptTimerBuffer.GetEnumerator();
					while(ide.MoveNext())
					{
						if(ide.Key.ToString().Equals(p_ID))
						{
							MyTimer myTimer = (MyTimer)ide.Value;
							myTimer.setEnabled(false);
							break;
						}
					}
				}
				if(ReceiptTimerBuffer.Contains(p_ID))
				{
					ReceiptTimerBuffer.Remove(p_ID);
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("Exception in ReleaseReceiptTimer(): "+e.Message+" ID: "+p_ID);
			}
		}

		private void Exit_Btn_Click(object sender, System.EventArgs e)
		{
			stop_server();			
			udpServerIsAlive = false;
			resetAll();
			this.Close();
		}

		private void SendFile_LB_SelectedIndexChanged(object sender, System.EventArgs e)
		{
		
		}

	}//end of main()

	public struct FilePacket
	{
		public string strType;
		public string strNumOfTotalPackets;
		public string strSenderIP;
		public string strID;
		public string strXMLFile;
	}

	public class MyTimer 
	{
		//properties	
		private Thread thread;
		private int _interval;
		private bool _enabled;
		private SendAndReceiveFiles _srf;
		private string _ID;
		private FilePacket _fp;
		
		public MyTimer(int p_interval, bool p_enabled, 
			SendAndReceiveFiles p_srf, string p_ID,FilePacket p_fp)
		{
			_interval = p_interval;
			_enabled = p_enabled;
			_srf = p_srf;
			_fp = p_fp;
			_ID = p_ID;
		}

		public void setEnabled(bool p_enabled)
		{
			_enabled = p_enabled;				 
		}

		public void setInterval(int p_interval)
		{
			_interval = p_interval;
		}

		public int getInterval()
		{
			return _interval;
		}

		public bool getEnabled()
		{
			return _enabled;
		}

		public void SetReceiptTimer()
		{			
			thread = new Thread(new ThreadStart(this.ReceiptTimer_Click));
			thread.Start();
		}

		private void ReceiptTimer_Click()
		{
			try
			{				
				while(_enabled)
				{				
					Monitor.Enter(this);
					Thread.Sleep(_interval);
					if(_enabled)
					{						
						_srf.ReleaseReceiptTimer(_ID);
						if(_srf.SenderBuffer.Contains(_ID))
						{
							_srf.SendFilePacket(_fp);
						}
						break;
					}												
					Monitor.Exit(this);
				}
			}
			catch(Exception e)
			{
				MessageBox.Show("ReceiptTimer_Click(): "+e.Message);
			}
		}
		
	}//inner class

}
